from statsmodels.api import GLM, OLS
from pandas import DataFrame
from estyp.linear_model.stepwise.__base import __both_selection, __forward_selection

def both_selection(formula: str, data: DataFrame, model: GLM, max_iter = 10000, **kwargs) -> str:
    """
# Both Forward and Backward Variable Selection for GLM's

This function performs both forward and backward variable selection using the Akaike Information Criterion (AIC).

## Parameters

* `formula`: A string representing the initial model formula.
* `data`: A Pandas DataFrame containing the data to be used for model fitting.
* `model`: A statsmodels.GLM object that represents the type of model to be fit.
* `max_iter`: The maximum number of iterations to perform.
* `**kwargs`: Additional keyword arguments to be passed to the model.from_formula() method.

## Returns

A string representing the final model formula.

## Example

```python
import statsmodels.api as sm
import pandas as pd

data = pd.DataFrame({
    "y": [1, 2, 3, 4, 5],
    "x1": [1, 2, 3, 4, 5],
    "x2": [6, 7, 8, 9, 10],
})

formula = "y ~ x1 + x2"

model = sm.OLS

final_formula = both_selection(formula=formula, data=data, model=model)

print(final_formula)
    """
    return __both_selection(formula, data, model, max_iter, **kwargs)

def forward_selection(
    y: str, 
    data: DataFrame,
    model: GLM = OLS, 
    alpha = 0.05,
    **kwargs
):
    """
# Forward Variable Selection for GLM's

This function performs forward variable selection using p-values calculated from nested models testing.

## Parameters

* `y`: A string containing the name of the dependent variable (target) to be predicted.
* data: The pandas DataFrame containing both the target variable 'y' and the        predictor variables for model training.
* `model`: A statsmodels model class. The statistical model to be used for model fitting and evaluation. Defaults to `sm.OLS`.
* `alpha`: A number between 0 and 1. The significance level for feature selection. A feature is added to the model if its p-value is less than this alpha value. Defaults to 0.05.
* `**kwargs`: Additional keyword arguments to be passed to the model.from_formula() method.

## Returns

A string representing the final model formula.

## Example

```python
import pandas as pd
import statsmodels.api as sm

# Create sample DataFrame
data = pd.DataFrame({
    'y': [1, 2, 3, 4, 5],
    'X1': [2, 4, 5, 7, 9],
    'X2': [3, 1, 6, 8, 4],
    'X3': [1, 5, 9, 2, 3]
})

# Perform the forward variable selection
formula = forward_selection(
    y = "y", ,
    data = data, 
    model = sm.OLS, 
    alpha = 0.05
)

# Fit the model using the selected formula
selected_model = sm.OLS.from_formula(formula, data).fit()
print(selected_model.summary())
```
    """
    return __forward_selection(y, data, model, alpha, **kwargs)


from typing import List, Literal, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from pandas import Series
from statsmodels.regression.linear_model import RegressionResultsWrapper

from estyp.testing.__base import (TestResults, _CheckHomocedasticity,
                                    _CheckIndependence,
                                    _CheckMulticollinearity, _CheckNormality,
                                    __nested_models_test, __prop_test, __t_test,
                                    __var_test)

class CheckModel:
    """
    # Check Linear Regression Assumptions

    The `CheckModel` class provides methods to test the assumptions of the linear regression model.
    These assumptions are:

    - Normality of residuals
    - Homoscedasticity (equal variance) of residuals
    - Independence of residuals
    - No Multicollinearity among predictors

    Parameters
    ----------
    `fitted_model` : `RegressionResultsWrapper`
        The fitted linear regression model which is an instance of RegressionResultsWrapper.

    Example
    -------
    ```python
    import statsmodels.api as sm
    from sklearn.datasets import load_diabetes
    diabetes = load_diabetes()
    X = diabetes["data"]
    y = diabetes["target"]
    X = sm.add_constant(X)
    model = sm.OLS(y, X)
    fitted_model = model.fit()
    cm = CheckModel(fitted_model)
    cm.check_all()
    ```
    """

    def __init__(self, fitted_model: RegressionResultsWrapper):
        if not isinstance(fitted_model, RegressionResultsWrapper):
            raise TypeError(
                "The fitted model must be a RegressionResultsWrapper object"
            )
        self.model = fitted_model

    def check_normality(self, alpha=0.05, plot=True, return_pvals=False):
        """
        Checks the normality assumption of the residuals using several statistical tests.

        Parameters
        ----------
        `alpha` : float, optional
            The significance level used by the statistical tests, by default 0.05.
        `plot` : bool, optional
            If `True`, a plot is shown for visual inspection of normality, by default `True`.
        `return_pvals` : bool, optional
            If `True`, the p-values of the statistical tests are returned, by default `False`.

        Returns
        -------
        dict
            A dictionary of p-values of the statistical tests if `return_pvals=True`.

        Example
        -------
        >>> cm.check_normality()
        """
        normality = _CheckNormality(self.model)
        if plot:
            fig, ax = normality.plot()
            plt.show()
        pvals = normality.test(sign_level=alpha)
        if return_pvals:
            return pvals

    def check_homocedasticity(self, alpha=0.05, plot=True, return_pvals=False):
        """
        Checks the homoscedasticity assumption (equal variance of residuals) using several statistical tests.
        
        Parameters
        ----------
        `alpha` : float, optional
            The significance level used by the statistical tests, by default 0.05.
        `plot` : bool, optional
            If `True`, a plot is shown for visual inspection of homoscedasticity, by default `True`.
        `return_pvals` : bool, optional
            If `True`, the p-values of the statistical tests are returned, by default `False`.
        
        Returns
        -------
        dict
            A dictionary of p-values of the statistical tests if `return_pvals=True`.
            
        Example
        -------
        >>> cm.check_homocedasticity()
        """
        homocedasticity = _CheckHomocedasticity(self.model)
        if plot:
            fig, ax = homocedasticity.plot()
            plt.show()
        pvals = homocedasticity.test(sign_level=alpha)
        if return_pvals:
            return pvals

    def check_independence(self, alpha=0.05, plot=True, return_vals=False):
        """
        Checks the independence assumption of the residuals using several statistical tests.
        
        Parameters
        ----------
        alpha : float, optional
            The significance level used by the statistical tests, by default 0.05.
        plot : bool, optional
            If `True`, a plot is shown for visual inspection of independence, by default `True`.
        return_pvals : bool, optional
            If `True`, the p-values of the statistical tests are returned, by default `False`.
        
        Returns
        -------
        dict
            A dictionary of values of the statistical tests if `return_vals=True`.
            
        Example
        -------
        >>> cm.check_independence()
        """
        independence = _CheckIndependence(self.model)
        if plot:
            fig, ax = independence.plot()
            plt.show()
        vals = independence.test(sign_level=alpha)
        if return_vals:
            return vals

    def check_multicollinearity(self, plot=True, return_cm=False):
        """
        Checks the multicollinearity assumption among predictors using the variance inflation factor (VIF).
        
        Parameters
        ----------
        `plot` : bool, optional
            If True, a plot is shown for visual inspection of multicollinearity, by default True.
        `return_cm` : bool, optional
            Returns the condition number of the model if True, by default False.
            
        Returns
        -------
        float
            The condition number of the model if `return_cm=True`.
        
        Example
        -------
        >>> cm.check_multicollinearity()
        """
        multicollinearity = _CheckMulticollinearity(self.model)
        if plot:
            fig, ax = multicollinearity.plot()
            plt.show()
        pvals = multicollinearity.test()
        if return_cm:
            return pvals

    def check_all(self, alpha=0.05, plot=True, return_vals=False):
        """
        Checks all the assumptions of the linear regression model.
        
        Parameters
        ----------
        `alpha` : float, optional
            The significance level used by the statistical tests, by default 0.05.
        `plot` : bool, optional
            If `True`, a plot is shown for visual inspection of each assumption, by default `True`.
        `return_vals` : bool, optional
            If `True`, the values of the statistical tests are returned, by default `False`.
        
        Returns
        -------
        dict
            A dictionary of values of the statistical tests if `return_vals=True`.
            
        Example
        -------
        >>> cm.check_all()
        """
        normality = _CheckNormality(self.model)
        homocedasticity = _CheckHomocedasticity(self.model)
        independence = _CheckIndependence(self.model)
        multicollinearity = _CheckMulticollinearity(self.model)

        normality_results = normality.test(sign_level=alpha)
        if plot:
            fig, ax = normality.plot()
            plt.show()
        homocedasticity_results = homocedasticity.test(sign_level=alpha)
        if plot:
            fig, ax = homocedasticity.plot()
            plt.show()
        independence_results = independence.test(sign_level=alpha)
        if plot:
            fig, ax = independence.plot()
            plt.show()
        multicollinearity_results = multicollinearity.test()
        if plot:
            fig, ax = multicollinearity.plot()
            plt.show()
        names = ["Normality", "Homocedasticity", "Independence", "Multicollinearity"]
        tests = [
            normality_results,
            homocedasticity_results,
            independence_results,
            multicollinearity_results,
        ]
        if return_vals:
            return dict(zip(names, tests))


def var_test(
    x: Union[List, np.ndarray, Series],
    y: Union[List, np.ndarray, Series],
    ratio: Union[float, int]=1,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    conf_level=0.95,
) -> TestResults:
    """
    # F Test to Compare Two Variances
    ## Description
    Performs an F test to compare the variances of two samples from normal populations. This function is inspired by the `var.test()` function of the software R.

    ## Usage

    ```python
    var_test(x, y, ...)
    var_test(x, y, ratio = 1,
            alternative = [two-sided", "less", "greater"],
            conf_level = 0.95)
    ```

    ## Arguments
    * `x`, `y`: numeric list, `np.array` or `pd.Series` of data values.
    * `ratio`: the hypothesized ratio of the population variances of x and y.
    * `alternative`: a character string specifying the alternative hypothesis, must be one of "two.sided" (default), "greater" or "less". You can specify just the initial letter.
    * `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.


    ## Details
    The null hypothesis is that the ratio of the variances of the populations from which x and y were drawn, is equal to ratio.

    ## Value
    An instance of the `TestResults` class containing the following attributes:

    * `statistic`: the value of the F test statistic.
    * `df`: the degrees of freedom for the F test statistic.
    * `p_value`: the p-value for the test.
    * `ci`: a confidence interval for the ratio of the population variances.
    * `estimate`: the ratio of the sample variances of x and y.
    * `alternative`: a string describing the alternative hypothesis.

    ## Examples

    ```python
    import numpy as np
    np.random.seed(2023)
    x = np.random.normal(size=100)
    y = np.random.normal(size=100)

    print(var_test(x, y))
    print(var_test(x, y, alternative="less"))
    print(var_test(x, y, ratio = 0.9, alternative="greater"))
    ```
    """
    return __var_test(x, y, ratio, alternative, conf_level)


def t_test(
    x: Union[List, np.ndarray, Series],
    y: Union[List, np.ndarray, Series] = None,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    mu: Union[float, int] = 0,
    paired=False,
    var_equal=False,
    conf_level=0.95,
) -> TestResults:
    """

    # Student's t-Test
    ## Description
    Performs one and two sample t-tests on groups of data. This function is inspired by the `t.test()` function of the software R.

    ## Usage
    ```python
    t_test(x, ...)
    ```

    ```python
    t_test(x, y = None,
        alternative = ["two-sided", "less", "greater"],
        mu = 0, paired = False, var_equal = False,
        conf_level = 0.95)
    ```


    ## Arguments
    * `x`: a (non-empty) numeric container of data values.
    * `y`: an (optional) numeric container of data values.
    * `alternative`: a string specifying the alternative hypothesis, must be one of "two-sided" (default), "greater" or "less".
    * `mu`: a number indicating the true value of the mean (or difference in means if you are performing a two sample test).
    * `paired`: a logical indicating whether you want a paired t-test.
    * `var_equal`: a logical variable indicating whether to treat the two variances as being equal. If `True` then the pooled variance is used to estimate the variance otherwise the Welch (or Satterthwaite) approximation to the degrees of freedom is used.
    * `conf_level`: a number between 0 and 1 indicanting the confidence level of the interval.


    ## Details
    alternative = "greater" is the alternative that x has a larger mean than y. For the one-sample case: that the mean is positive.

    If `paired` is `True` then both x and y must be specified and they must be the same length. Missing values are silently removed (in pairs if `paired` is `True`). If `var_equal` is `True` then the pooled estimate of the variance is used. By default, if `var_equal` is `False` then the variance is estimated separately for both groups and the Welch modification to the degrees of freedom is used.


    ### Value
    An instance of the `TestResults` class containing the following attributes:

    * `statistic`: the value of the t-statistic.
    * `df`: the degrees of freedom for the t-statistic.
    * `p_value`: the p-value for the test.
    * `ci`: a confidence interval for the mean appropriate to the specified alternative hypothesis.
    * `estimate`: the estimated mean or list of estimated means depending on whether it was a one-sample test or a two-sample test.
    * `alternative`: a character string describing the alternative hypothesis.
    * `mu`: the mean of the null hypothesis.


    ## Examples
    ```python
    import numpy as np
    np.random.seed(2023)
    x = np.random.normal(size=100)
    y = np.random.normal(size=100)
    mu = 0.1

    print(t_test(x, mu=mu, alternative="less"))
    print(t_test(x, y, mu=mu))
    print(t_test(x, y, mu=mu, var_equal=True, alternative="greater"))
    print(t_test(x, y, mu=mu, paired=True))
    ```
    """
    return __t_test(x, y, alternative, mu, paired, var_equal, conf_level)


def nested_models_test(
    fitted_small_model: RegressionResultsWrapper,
    fitted_big_model: RegressionResultsWrapper,
) -> TestResults:
    """
    # Nested Models F-Test Function

    ## Description

    This function performs a nested models F-test using deviance from two fitted models from statsmodels library. The test compares two nested models: a larger or "big" model and a smaller or "small" model. The purpose of this test is to determine whether the larger model significantly improves the model fit compared to the smaller model by adding additional predictors.

    ## Parameters
    * `fitted_small_model`: The fitted model representing the smaller/nested model. It have to come from statsmodels.
    * `fitted_big_model`: : The fitted model representing the larger model, which includes all the predictors from the smaller model and potentially additional predictors. It have to come from statsmodels.

    ## Returns

    The function returns an object of class TestResults that contains the following information:

    * `method`: A string indicating the name of the statistical test (Nested models F-test).
    * `statistic`: The computed F-statistic value.
    * `estimate`: The difference in deviances between the models.
    * `df`: A dictionary with the degrees of freedom for the numerator and denominator of the F-statistic.
    * p_value: The p-value associated with the F-statistic.

    ## Examples
    ### Example 1: With OLS
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 5.01, 6.89],
        "y": [2, 3, 4, 5, 6]
    })

    model_small = sm.OLS.from_formula("y ~ 1", data).fit()
    model_big = sm.OLS.from_formula("y ~ x", data).fit()

    print(nested_models_test(model_small, model_big))
    ```
    ### Example 2: With Logit
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 3.01, 4.89],
        "y": [0, 1, 1, 0, 1]
    })

    model_small = sm.Logit.from_formula("y ~ 1", data).fit()
    model_big = sm.Logit.from_formula("y ~ x", data).fit()

    print(nested_models_test(model_small, model_big))
    ```
    ### Example 3: With GLM
    ```python
    import pandas as pd
    import statsmodels.api as sm
    data = pd.DataFrame({
        "x": [2.01, 2.99, 4.01, 5.01, 6.89],
        "y": [2, 3, 4, 5, 6]
    })
    model_small = sm.GLM.from_formula("y ~ 1", data, family = sm.families.Gamma()).fit()
    model_big = sm.GLM.from_formula("y ~ x", data, family = sm.families.Gamma()).fit()

    print(nested_models_test(model_small, model_big))
    ```
    """
    return __nested_models_test(fitted_small_model, fitted_big_model)


def prop_test(
    x: Union[List, np.ndarray, Series],
    n: Optional[Union[List, np.ndarray, Series]] = None,
    p: Optional[Union[float, List, np.ndarray]] = None,
    alternative: str = "two-sided",
    conf_level: float = 0.95,
    correct: bool = True,
) -> TestResults:
    """
    # Test of Equal or Given Proportions.

    `prop_test()` can be used for testing the null that the proportions
    (probabilities of success) in several groups are the same, or that they equal
    certain given values.

    Parameters
    ----------
    `x` : array_like
        a vector of counts of successes, a one-dimensional table with two entries,
        or a two-dimensional table (or matrix) with 2 columns, giving the counts of
        successes and failures, respectively.
    `n` : array_like, optional
        a vector of counts of trials; ignored if x is a matrix or a table.
        If not provided, it is calculated as the sum of the elements in x.
    `p` : array_like, optional
        a vector of probabilities of success. The length of p must be the same as
        the number of groups specified by x, and its elements must be greater than
        0 and less than 1.
    alternative : str, optional
        a character string specifying the alternative hypothesis, must be one of
        "two.sided" (default), "greater" or "less". You can specify just the
        initial letter. Only used for testing the null that a single proportion
        equals a given value, or that two proportions are equal; ignored otherwise.
    `conf_level` : float, optional
        confidence level of the returned confidence interval. Must be a single
        number between 0 and 1. Only used when testing the null that a single
        proportion equals a given value, or that two proportions are equal;
        ignored otherwise.
    `correct` : bool, optional
        a logical indicating whether Yates' continuity correction should be
        applied where possible.

    Returns
    -------
    `TestResult`
        A data class with the following attributes:

        `statistic` : float
            the value of Pearson's chi-squared test statistic.
        `df` : int
            the degrees of freedom of the approximate chi-squared distribution of
            the test statistic.
        `p_value` : float
            the p-value of the test.
        `estimate` : array_like
            a vector with the sample proportions x/n.
        `null_value` : float or array_like
            the value of p if specified by the null hypothesis.
        `conf_int` : array_like
            a confidence interval for the true proportion if there is one group,
            or for the difference in proportions if there are 2 groups and p is
            not given, or None otherwise. In the cases where it is not None, the
            returned confidence interval has an asymptotic confidence level as
            specified by conf_level, and is appropriate to the specified
            alternative hypothesis.
        `alternative` : str
            a character string describing the alternative.
        `method` : str
            a character string indicating the method used, and whether Yates'
            continuity correction was applied.

    Example
    --------
    ```python
    import numpy as np
    from scipy import stats
    x = np.array([83, 90, 129, 70])
    n = np.array([86, 93, 136, 82])
    result = prop_test(x, n)
    print(result)
    ```
    """
    return __prop_test(x, n, p, alternative, conf_level, correct)




import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm
from pandas import DataFrame, Series
from scipy.stats import chi2
from scipy.stats import f as fisher
from scipy.stats import (kstest, norm, probplot, shapiro, ttest_1samp,
                         ttest_ind, ttest_rel)
from statsmodels.api import GLM, OLS
from statsmodels.genmod.families.family import Gaussian
from statsmodels.stats.diagnostic import (acorr_breusch_godfrey,
                                          acorr_ljungbox, het_breuschpagan,
                                          het_goldfeldquandt, het_white)
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.stats.stattools import (durbin_watson, jarque_bera,
                                         omni_normtest)


class bcolors:
    HEADER = "\033[95m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    
class _CheckNormality:
    def __init__(self, fitted_model):
        self.model = fitted_model
        self.__e = fitted_model.resid
        self.__scale = fitted_model.scale ** (1 / 2)
        self.__norm = norm(loc=0, scale=self.__scale)

    def plot(self):
        xn = np.linspace(self.__e.min() * 1.3, self.__e.max() * 1.3, 100)
        yn = self.__norm.pdf(xn)
        x, y = probplot(self.__e, fit=False, dist=self.__norm)
        minimo = self.__e.min() * 1.2
        maximo = self.__e.max() * 1.2

        # Graficos
        # G1
        fig, ax = plt.subplots(figsize=(12, 4), ncols=2, nrows=1)
        ax[0].hist(self.__e, bins=20, edgecolor="white", density=True, alpha=0.9)
        ax[0].plot(xn, yn, color="darkgreen")
        ax[0].fill_between(xn, 0, yn, color="darkgreen", alpha=0.3)
        # G2
        ax[1].scatter(x, y, alpha=0.8)
        ax[1].plot((-1e10, 1e10), (-1e10, 1e10), color="darkgreen", linewidth=2)
        ax[1].set_xlim(minimo, maximo)
        ax[1].set_ylim(minimo, maximo)

        # Etiquetas
        fig.suptitle("Normality of residuals:")
        # G1
        ax[0].set_xlabel("Residuals")
        ax[0].set_ylabel("Density")
        ax[0].set_title("Histogram should be similar to the curve", fontsize=10)
        # G2
        ax[1].set_ylabel("Residuals Quantiles")
        ax[1].set_xlabel("Theoretical Quantiles")
        ax[1].set_title("Points should be close to the line", fontsize=10)
        return fig, ax

    def test(self, sign_level=0.05):
        _, pval_ks = kstest(self.__e, cdf=self.__norm.cdf)
        _, pval_sh = shapiro(self.__e)
        _, pval_jb, *_ = jarque_bera(self.__e)
        _, pval_om = omni_normtest(self.__e)

        text_ok = (
            lambda name, p: f"- Residuals appear as normally distributed according to {name} test (p-value = {p:0.3f})."
        )
        test_no = (
            lambda name, p: f"- Residuals don't appear as normally distributed according to {name} test (p-value = {p:0.3f})."
        )
        names = ["KS", "Shapiro-Wilk", "Jarque-Bera", "Omni"]
        pvals = [pval_ks, pval_sh, pval_jb, pval_om]

        print(
            bcolors.BOLD + bcolors.UNDERLINE + "Normality tests results:" + bcolors.ENDC
        )

        for name, pval in zip(names, pvals):
            if pval >= sign_level:
                print(bcolors.OKGREEN + text_ok(name, pval) + bcolors.ENDC)
            else:
                print(bcolors.FAIL + test_no(name, pval) + bcolors.ENDC)
        return dict(zip(names, pvals))


class _CheckHomocedasticity:
    def __init__(self, fitted_model):
        self.model = fitted_model
        self.__e = fitted_model.resid

    def plot(self):
        std_resid = self.model.get_influence().resid_studentized_internal
        fig, ax = plt.subplots(figsize=(6, 4))
        ax.scatter(self.model.fittedvalues, std_resid, alpha=0.8)
        ax.axhline(0, color="darkgreen", linestyle="--")
        ax.set_xlabel("Fitted Values")
        ax.set_ylabel("Studentized Residuals")
        ax.set_title(
            "Homocedasticity of residuals:\nIt shouldn't be any pattern in the plot",
            fontsize=10,
        )
        return fig, ax

    def test(self, sign_level=0.05):
        *_, pval_bp = het_breuschpagan(self.__e, self.model.model.exog)
        *_, pval_w = het_white(self.__e, self.model.model.exog)
        _, pval_gq, _ = het_goldfeldquandt(self.__e, self.model.model.exog)
        text_ok = (
            lambda name, p: f"- Error variance appears to be homoscedastic according to {name} test (p-value = {p:0.3f})."
        )
        test_no = (
            lambda name, p: f"- Heteroscedasticity (non-constant error variance) detected according to {name} test (p-value = {p:0.3f})."
        )
        names = ["Breusch-Pagan", "White", "Goldfeld-Quandt"]
        pvals = [pval_bp, pval_w, pval_gq]
        print(
            bcolors.BOLD
            + bcolors.UNDERLINE
            + "Homocedasticity tests results:"
            + bcolors.ENDC
        )

        for name, pval in zip(names, pvals):
            if pval >= sign_level:
                print(bcolors.OKGREEN + text_ok(name, pval) + bcolors.ENDC)
            else:
                print(bcolors.FAIL + test_no(name, pval) + bcolors.ENDC)
        return dict(zip(names, pvals))


class _CheckIndependence:
    def __init__(self, fitted_model):
        self.model = fitted_model
        self.__e = fitted_model.resid

    def plot(self):
        fig, ax = plt.subplots(figsize=(6, 4))
        sm.graphics.tsa.plot_acf(self.__e, lags=20, ax=ax)
        ax.set_xlabel("Lag")
        ax.set_ylabel("Autocorrelation")
        ax.set_title(
            "Independence of residuals:\nAll bars should be within the blue region",
            fontsize=10,
        )
        return fig, ax

    def test(self, sign_level=0.05):
        dw_stat = durbin_watson(self.__e)
        pval_bp = acorr_ljungbox(self.__e, boxpierce=True, auto_lag=True)[
            "bp_pvalue"
        ].min()
        *_, pval_bg = acorr_breusch_godfrey(self.model)
        text_ok = (
            lambda name, p: f"- Residuals appear to be independent and not autocorrelated according to {name} test (p-value = {p:0.3f})."
        )
        test_no = (
            lambda name, p: f"- Autocorrelated residuals detected according to {name} test (p-value = {p:0.3f})."
        )
        names = ["Durbin-Watson Stat", "Box-Pierce", "Breusch-Godfrey"]
        pvals = [dw_stat, pval_bp, pval_bg]

        print(
            bcolors.BOLD
            + bcolors.UNDERLINE
            + "Independence tests results:"
            + bcolors.ENDC
        )

        if 1.5 <= dw_stat <= 2.5:
            print(
                bcolors.OKGREEN
                + f"- Residuals appear to be independent and not autocorrelated according to DW test (DW-Statistic = {dw_stat:0.3f})"
                + bcolors.ENDC
            )
        else:
            print(
                bcolors.FAIL
                + f"- Autocorrelated residuals detected according to DW test (DW-Statistic = {dw_stat:0.3f})"
                + bcolors.ENDC
            )

        for name, pval in zip(names[1:], pvals[1:]):
            if pval >= sign_level:
                print(bcolors.OKGREEN + text_ok(name, pval) + bcolors.ENDC)
            else:
                print(bcolors.FAIL + test_no(name, pval) + bcolors.ENDC)
        return dict(zip(names, pvals))


class _CheckMulticollinearity:
    def __init__(self, fitted_model):
        self.fitted_model = fitted_model
        self.exog_df = DataFrame(
            fitted_model.model.exog, columns=fitted_model.model.exog_names
        )

    def test(self):
        cn = self.fitted_model.condition_number
        print(
            bcolors.BOLD
            + bcolors.UNDERLINE
            + "Multicollinearity test results:"
            + bcolors.ENDC
        )
        if cn < 20:
            print(
                bcolors.OKGREEN
                + f"The model is unlikely to have multicollinearity problems (condition number = {cn:,.2f})."
                + bcolors.ENDC
            )
        else:
            print(
                bcolors.FAIL
                + f"- The model may have multicollinearity problems (condition number = {cn:,.2f})."
                + bcolors.ENDC
            )

    def plot(self):
        # Calcula el VIF para cada predictor
        vif = DataFrame()
        vif["VIF Factor"] = [
            variance_inflation_factor(self.exog_df.values, i)
            for i in range(self.exog_df.shape[1])
        ]
        vif["features"] = self.exog_df.columns

        # Ordena el dataframe por el VIF
        vif = vif.sort_values(by="VIF Factor")

        # Dibuja el gráfico de barras
        fig, ax = plt.subplots(figsize=(10, 6))
        colors = [
            "dodgerblue" if 0 < x <= 5 else ("green" if 5 < x < 10 else "red")
            for x in vif["VIF Factor"]
        ]
        ax.hlines(y=vif["features"], xmin=0, xmax=vif["VIF Factor"], color=colors)
        for x, y, color in zip(vif["VIF Factor"], vif["features"], colors):
            ax.plot(x, y, "o", color=color)
        for x, y, tex in zip(vif["VIF Factor"], vif["features"], vif["VIF Factor"]):
            t = ax.text(
                x=x + 0.05 if max(vif["VIF Factor"]) < 7 else x + 1,
                y=y,
                s=round(tex, 2),
                horizontalalignment="left",
                verticalalignment="center",
                fontdict={
                    "color": "dodgerblue"
                    if 0 < x <= 5
                    else ("green" if 5 < x < 10 else "red"),
                    "size": 14,
                },
            )

        ax.fill_betweenx(vif["features"], vif["VIF Factor"], color="gray", alpha=0.1)

        ax.set_xlabel("VIF Factor")
        ax.set_title(
            "Variance Inflation Factor (VIF):\nAll bars should be below 10", size=10
        )
        ax.spines[["right", "top"]].set_visible(False)
        return fig, ax


class TestResults:
    def __init__(self, res: dict, names: dict):
        self.__dict__.update(res)
        self.__names = names

    def __repr__(self):
        # p-value
        p_value = self.p_value
        if p_value < 0.0001:
            p_value = "<0.0001"
        else:
            p_value = f"{p_value:0.4f}"
        # df
        df = self.df
        if isinstance(df, (float, int)):
            if df == float(int(df)):
                df = int(df)
        elif isinstance(df, dict):
            df = {k: f"{v:0.2f}" for k, v in df.items()}
        elif isinstance(df, float):
            df = f"{df:0.2f}"
        # estimates
        estimate = self.estimate
        if isinstance(estimate, (list, np.ndarray)):
            estimate = [float(f"{e:0.6f}") for e in estimate]
        elif estimate == float(int(estimate)):
            estimate = int(estimate)
        elif isinstance(estimate, float):
            estimate = f"{estimate:0.6f}"

        string = f"""
    {self.method}
    {len(self.method) * "-"}
    {self.__names['statistic']} = {self.statistic:0.4f} | df: {df} | p-value = {p_value}
    alternative hypothesis: {self.__names["alternative"]}"""
        if self.__dict__.get("conf_int"):
            if self.conf_int is not None:
                cl = self.conf_level * 100
                if cl == float(int(cl)):
                    cl = int(cl)
                string += f"""
        {cl} percent confidence interval:
        {" "}{self.conf_int[0]:0.6f} {self.conf_int[1]:0.6f}"""
        string += f"""
    sample estimates:
    {" " * 2}{self.__names["estimate"]}: {estimate}
    """
        return string


def __var_test(x, y, ratio, alternative, conf_level):
    if not (isinstance(ratio, float) or isinstance(ratio, int)):
        raise ValueError("'ratio' must be a single positive number")
    else:
        if not ratio > 0:
            raise ValueError("'ratio' must be a single positive number")

    if alternative not in ["two-sided", "less", "greater"]:
        raise ValueError(
            "'alternative' must be one of 'two.sided', 'less', or 'greater'"
        )

    if not (isinstance(conf_level, float)):
        raise ValueError("'conf_level' must be a single number between 0 and 1")
    else:
        if not (conf_level > 0 and conf_level < 1):
            raise ValueError("'conf_level' must be a single number between 0 and 1")

    if isinstance(x, np.ndarray) or isinstance(x, Series):
        df_x = x.shape[0] - 1
        v_x = np.var(x)
    else:
        df_x = len(x) - 1
        v_x = np.var(x)

    if isinstance(y, np.ndarray) or isinstance(y, Series):
        df_y = y.shape[0] - 1
        v_y = np.var(y)
    else:
        df_y = len(y) - 1
        v_y = np.var(y)

    est = v_x / v_y
    stat = est / ratio

    f = fisher(dfn=df_x, dfd=df_y)
    pval = f.cdf(stat)

    if alternative == "two-sided":
        pval = 2 * min([pval, 1 - pval])
        beta = (1 - conf_level) / 2
        ci = [est / f.ppf(1 - beta), est / f.ppf(beta)]
    elif alternative == "greater":
        pval = 1 - pval
        ci = [est / f.ppf(conf_level), np.inf]
    else:
        ci = [0, est / f.ppf(1 - conf_level)]

    alternative_name = (
        "not equal to" if alternative == "two-sided" else f"{alternative} than"
    )

    names = {
        "statistic": "F",
        "estimate": "ratio of variances",
        "alternative": f"true ratio of variances is {alternative_name} {ratio}",
    }

    res = {
        "method": "F test to compare two variances",
        "statistic": stat,
        "estimate": est,
        "df": {"x": df_x, "y": df_y},
        "p_value": pval,
        "conf_int": ci,
        "conf_level": conf_level,
        "alternative": alternative,
    }

    return TestResults(res, names)


def __t_test(x, y, alternative, mu, paired, var_equal, conf_level):
    if isinstance(conf_level, float):
        if conf_level <= 0 or conf_level >= 1:
            raise ValueError("'conf_level' must be between 0 and 1")
    else:
        raise TypeError("'conf_level' must be a single number between 0 and 1")

    if isinstance(alternative, str):
        if alternative not in ["two-sided", "less", "greater"]:
            raise ValueError(
                "'alternative' must be one of 'two-sided', 'less', or 'greater'"
            )
    else:
        raise TypeError("'alternative' must be a string")

    if paired:
        if y is None:
            raise ValueError("If 'paired' is True, 'y' must be provided")
        else:
            result = ttest_rel(x, y + mu, alternative=alternative)
    else:
        if y is None:
            result = ttest_1samp(x, mu, alternative=alternative)
        else:
            result = ttest_ind(x, y + mu, alternative=alternative, equal_var=var_equal)

    lower, upper = result.confidence_interval(conf_level)
    ci = (lower, upper)
    t, p = result.statistic, result.pvalue
    df = result.df

    alternative_dire = (
        "not equal to" if alternative == "two-sided" else f"{alternative} than"
    )
    alternative_conc = "mean difference" if paired else "difference in means"

    if y is None:
        est = x.mean()
        estimate = "mean of x"
        alternative_name = f"true mean is {alternative_dire} {mu}"
        method = "One Sample t-test"
    else:
        welch = "Welch's " if (not var_equal and not paired) else ""
        title = "Paired" if paired else "Two Sample"
        est = [x.mean(), y.mean()]
        estimate = "[mean of x, mean of y]"
        alternative_name = f"true {alternative_conc} is {alternative_dire} {mu}"
        method = f"{welch}{title} t-test"

    names = {
        "statistic": "T",
        "estimate": estimate,
        "alternative": alternative_name,
    }

    res = {
        "method": method,
        "statistic": t,
        "estimate": est,
        "df": df,
        "p_value": p,
        "conf_int": ci,
        "mu": mu,
        "conf_level": conf_level,
        "alternative": alternative,
    }

    return TestResults(res, names)


def __deviance(fitted_model, response):
    if isinstance(fitted_model.model, OLS):
        return Gaussian().deviance(endog=response, mu=fitted_model.fittedvalues)
    elif isinstance(fitted_model.model, GLM):
        return fitted_model.deviance
    return sum(fitted_model.resid_dev ** 2) 


def __nested_models_test(fitted_small_model, fitted_big_model):
    response = fitted_big_model.model.data.endog
    n = fitted_big_model.fittedvalues.shape[0]
    p_big = fitted_big_model.params.shape[0] - 1
    p_small = fitted_small_model.params.shape[0] - 1
    d_big = __deviance(fitted_big_model, response)
    d_small = __deviance(fitted_small_model, response)
    df_num = p_big - p_small
    df_den = n - p_big - 1
    f_num = (d_small - d_big) / df_num
    f_den = d_big / df_den
    f_stat = f_num / f_den
    p_value = fisher(dfn=df_num, dfd=df_den).sf(f_stat)
    names = {
        "statistic": "F",
        "estimate": "Difference in deviances between models",
        "alternative": "big model is true",
    }

    res = {
        "method": "Nested models F-test",
        "statistic": f_stat,
        "estimate": d_small - d_big,
        "df": {"df_num": df_num, "df_den": df_den},
        "p_value": p_value,
    }

    return TestResults(res, names)


def __prop_test(x, n, p, alternative, conf_level, correct):
    p_status = p is not None
    if isinstance(x, list):
        x = np.array(x)
    if isinstance(x, (Series, DataFrame)):
        x = x.values
    if isinstance(x, (int, float)):
        x = np.array([x])

    
    if n is None and len(x.shape) == 1:
        if x.shape[0] != 2:
            raise ValueError("'x' should have 2 entries")
        l = 1
        n = np.sum(x)
        x = np.array(x[0])
    elif len(x.shape) == 2:  # If x is a matrix
        if x.shape[1] != 2:
            raise ValueError("'x' must have 2 columns")
        l = x.shape[0]
        n = np.sum(x, axis=1)
        x = x[:, 0]
    elif n is not None and not isinstance(n, int) and len(x) != len(n):
        l = x.shape[0]
        raise ValueError("'x' and 'n' must have the same length")

    # Ensure finite data
    x = np.array(x)[np.isfinite(x)]
    n = np.array(n)[np.isfinite(n)]

    k = x.shape[0]
    if k < 1:
        raise ValueError("Not enough data")

    if any(n <= 0):
        raise ValueError("Elements of 'n' must be positive")

    if any(x < 0):
        raise ValueError("Elements of 'x' must be nonnegative")

    if any(x > n):
        raise ValueError("Elements of 'x' must not be greater than those of 'n'")

    if p is None and k == 1:
        p = np.array([0.5])
    
    if p is not None:
        if isinstance(p, list):
            p = np.array(p)
        else:
            p = np.array([p])

        if len(p) != k:
            raise ValueError("'p' must have the same length as 'x' and 'n'")

        if any(np.array(p) <= 0) or any(np.array(p) >= 1):
            raise ValueError("Elements of 'p' must be in (0,1)")

    if k > 2 or (k == 2 and p is not None):
        alternative = "two-sided"

    if not (0 < conf_level < 1):
        raise ValueError("'conf_level' must be a single number between 0 and 1")

    if isinstance(n, (list, Series)):
        n = np.array(n)

    estimate = x / n
    yates = 0.5 if correct and k <= 2 else 0
    if k == 1:
        z = norm.ppf((1 + conf_level) / 2) if alternative == "two-sided" else norm.ppf(conf_level)
        yates = min(yates, abs(x[0] - n * p))
        z22n = z**2 / (2 * n)
        p_c = estimate[0] + yates / n
        p_u = 1 if p_c >= 1 else (p_c + z22n + z * np.sqrt(p_c * (1 - p_c) / n + z22n / (2 * n))) / (1 + 2 * z22n)
        p_c = estimate[0] - yates / n
        p_l = 0 if p_c <= 0 else (p_c + z22n - z * np.sqrt(p_c * (1 - p_c) / n + z22n / (2 * n))) / (1 + 2 * z22n)
        conf_int = {
            "two-sided": (max(p_l.item(), 0), min(p_u.item(), 1)),
            "greater": (max(p_l.item(), 0), 1),
            "less": (0, min(p_u.item(), 1))
        }[alternative]
        conf_int = (conf_int[0], conf_int[1])
    elif k == 2 and p is None:
        delta = estimate[0] - estimate[1]
        yates = min(yates, abs(delta) / np.sum(1 / n))
        width = (norm.ppf((1 + conf_level) / 2) if alternative == "two-sided" else norm.ppf(conf_level)) * \
                np.sqrt(np.sum(estimate * (1 - estimate) / n)) + yates * np.sum(1 / n)
        conf_int = {
            "two-sided": (max(delta - width, -1), min(delta + width, 1)),
            "greater": (max(delta - width, -1), 1),
            "less": (-1, min(delta + width, 1))
        }[alternative]
    else:
        conf_int = None

    if p is None:
        p = np.sum(x) / np.sum(n)
        parameter = k - 1
    else:
        parameter = k
    e = np.vstack([n * p, n * (1 - p)]).T
    X = np.vstack([x, n - x]).T
    chi2_stat = np.sum((np.abs(X - e) - yates)**2 / e)

    if alternative == "two-sided":
        p_value = 1 - chi2.cdf(chi2_stat, parameter)
    else:
        z = np.sign(estimate - p) * np.sqrt(chi2_stat)
        if isinstance(z, np.ndarray):
            if z.shape[0] > 1:
                z = z[0]
        p_value = 1 - norm.cdf(z) if alternative == "greater" else norm.cdf(z)

    
    method = f"{k}-sample test for {'equality of' if p_status else 'given'} proportions {'with' if yates else 'without'} continuity correction"

    
    direction = "not equal to" if alternative == "two-sided" else f"{alternative} than"
    if not p_status:  # if p is None
        if k == 1:
            alternative_name = f"the true proportion is {direction} 0.5"
        else:
            alternative_name = "the true proportions are not all equal"
    else:  # if p is not None
        if k == 1:
            alternative_name = f"the true proportion is {direction} {p[0]:0.4f}"
        else:
            alternative_name = f"the true proportions are {direction} {p:0.4f}"
    
    names = {
        "statistic": "X-squared",
        "estimate": "proportion(s)",
        "alternative": alternative_name,
    }
    res = {
        "method": method,
        "statistic": chi2_stat,
        "estimate": estimate.item() if k == 1 else estimate,
        "df": parameter,
        "p_value": p_value.item() if alternative != "two-sided" else p_value,
        "conf_int": conf_int,
        "null_value": p.item() if len(p.shape) != 0 and p.shape[0] == 1 else p,
        "conf_level": conf_level,
        "alternative": alternative,
    }
    #return res
    return TestResults(res, names)